import tempfile
import os

from requests import HTTPError

from test_pyczds import TestPyCZDS


class MockRequest(object):
    def __init__(self, status_code=200, headers='', text=''):
        self.status_code = status_code
        self.headers = headers
        self.text = text


class TestHelpers(TestPyCZDS):
    # region offline tests
    def test_http_status_400_offline(self):
        r = MockRequest(400)
        with self.assertRaisesRegex(HTTPError, '.malformed.'):
            self.client._preprocess_response(r)

    def test_http_status_401_offline(self):
        r = MockRequest(401)
        with self.assertRaisesRegex(HTTPError, '.nvalid credentials.*no valid bearer token.'):
            self.client._preprocess_response(r)

    def test_http_status_403_offline(self):
        r = MockRequest(403)
        with self.assertRaisesRegex(HTTPError, '.*zone.*not exist.*not have access*'):
            self.client._preprocess_response(r)

    def test_http_status_409_offline(self):
        r = MockRequest(409)
        with self.assertRaisesRegex(HTTPError, '.T\&Cs.'):
            self.client._preprocess_response(r)

    def test_http_status_415_offline(self):
        r = MockRequest(415)
        with self.assertRaisesRegex(HTTPError, '.nsupported content type.'):
            self.client._preprocess_response(r)

    def test_http_status_429_offline(self):
        r = MockRequest(429)
        with self.assertRaisesRegex(HTTPError, '.oo many authentication attempts.'):
            self.client._preprocess_response(r)

    def test_http_status_500_offline(self):
        r = MockRequest(500)
        with self.assertRaisesRegex(HTTPError, '.nternal service error.'):
            self.client._preprocess_response(r)

    def test_file_size_difference(self):
        test_string = '1234567890'
        test_string_size = len(test_string) - 1

        with tempfile.TemporaryDirectory() as d:
            file_path = os.path.join(d, 'file_size_test.tmp')
            with open(file_path, 'w') as f:
                f.write(test_string)

            with self.assertRaisesRegex(Exception, '.*size of the file.*differs.*'):
                self.client._check_file_size(file_path, test_string_size)

    def test_file_size_identical(self):
        test_string = '1234567890'
        test_string_size = len(test_string)

        with tempfile.TemporaryDirectory() as d:
            file_path = os.path.join(d, 'file_size_test.tmp')
            with open(file_path, 'w') as f:
                f.write(test_string)

            self.client._check_file_size(file_path, test_string_size)

    # endregion



import sys
import logging
import os
from urllib.parse import urljoin, urlparse, unquote
from email.utils import parsedate_to_datetime
from email.message import EmailMessage

import requests

from .authentication import CZDSAuthentication


class CZDSClient(CZDSAuthentication):
    BASE_URL = 'https://czds-api.icann.org'
    ZONE_DOWNLOAD_LINKS_LIST_URL = urljoin(BASE_URL, '/czds/downloads/links')

    def __init__(self, username: str, password: str) -> None:
        super().__init__(username=username, password=password)

        self._user_agent = '{} / {}'.format('pyCZDS', '1.6')

        if 'unittest' in sys.modules.keys():
            self._user_agent = '{} TEST'.format(self._user_agent)

    def _do_request(self, method: str, url: str, stream: bool = False) -> requests.Response:
        if not self._is_authenticated():
            logging.debug('Not authenticated. Attempting authentication.')
            self._authenticate()

        headers = {
            'Authorization': 'Bearer {}'.format(self._token),
            'User-Agent': self._user_agent
        }

        request = requests.Request(method.upper(), url=url, headers=headers)
        prepared_request = request.prepare()
        self._preprocess_request(prepared_request)

        response = requests.Session().send(prepared_request, stream=stream)
        self._preprocess_response(response, stream)

        return response

    def _parse_headers(self, headers: requests.models.CaseInsensitiveDict) -> requests.models.CaseInsensitiveDict:
        parsed = requests.models.CaseInsensitiveDict()

        parsed['last-modified'] = parsedate_to_datetime(headers['last-modified'])
        parsed['content-length'] = int(headers['content-length'])

        msg = EmailMessage()
        msg['content-disposition'] = headers['content-disposition']
        filename = msg.get_filename()
        parsed['filename'] = filename

        headers['parsed'] = parsed

        return headers

    def get_zonefiles_list(self) -> list:
        logging.debug('About to request zonefile URLs list.')

        url_list = self._do_request('get', self.ZONE_DOWNLOAD_LINKS_LIST_URL).json()

        if len(url_list) == 0:
            raise Exception('This account does not seem to have access to any zonefiles.')

        return url_list

    def head_zonefile(self, zonefile_url: str) -> requests.models.CaseInsensitiveDict:
        logging.debug('About to request headers for zonefile {}.'.format(zonefile_url.split('/')[-1]))

        headers = self._do_request('head', zonefile_url).headers
        headers = self._parse_headers(headers)

        return headers

    def get_zonefile(self, zonefile_url: str, download_dir: str = '', filename: str = '') -> None:
        if not download_dir or (download_dir and len(str(download_dir)) == 0):
            download_dir = os.getcwd()

        url_parsed = urlparse(zonefile_url)
        remote_filename = unquote(os.path.basename(url_parsed.path))

        logging.debug('About to start download for zonefile {} to directory {}.'.format(
            remote_filename, download_dir)
        )

        with self._do_request('get', zonefile_url, stream=True) as response:
            headers = self._parse_headers(response.headers)
            header_size = headers['parsed']['content-length']

            if not filename or (filename and len(filename) == 0):
                local_filename = headers['parsed']['filename']
            else:
                local_filename = filename

            file_path = os.path.join(download_dir, local_filename)

            logging.debug('Streaming file with {:,} bytes to {}.'.format(header_size, file_path))

            with open(file_path, 'wb') as f:
                for chunk in response.iter_content(chunk_size=2**19):
                    f.write(chunk)
                    file_size = os.path.getsize(file_path)
                    logging.info(
                        '[{:.0%}] Downloaded {:,} of {:,} bytes.'.format(
                            file_size / header_size, file_size, header_size
                        )
                    )

            logging.debug('Completed download of zonefile {}.'.format(file_path))

            self._check_file_size(file_path, header_size)



import logging
import json
import base64
import time

import requests

from .helpers import CZDSHelpers


class CZDSAuthentication(CZDSHelpers):
    AUTH_URL = "https://account-api.icann.org/api/authenticate"

    def __init__(self, username: str, password: str) -> None:
        if not isinstance(username, str) or len(username) == 0 or not self._is_email_address(username):
            raise ValueError('Username invalid.')

        if not isinstance(password, str) or len(password) == 0:
            raise ValueError('Password invalid.')

        self._username = username
        self._password = password

        self._token = str()

    def _get_token_jwt_payload(self) -> json:
        logging.debug('Parsing JWT payload.')

        if len(self._token) == 0:
            raise ValueError('Authentication token is not set.')

        payload_b64 = self._token.split('.')[1]
        s = base64.b64decode(payload_b64 + '=' * (-len(payload_b64) % 4))
        j = json.loads(s)

        logging.debug('JSON content of JWT payload: {}'.format(json.dumps(j)))

        return j

    def _is_authenticated(self) -> bool:
        logging.debug('About to check if client is authenticated or not.')

        if len(self._token) > 0:
            j = self._get_token_jwt_payload()

            if j['exp'] > time.time():
                logging.debug('Token is set and has not yet expired. Client is authenticated.')
                return True

            logging.debug('Not authenticated because token has expired.')

        else:
            logging.debug('Not authenticated because auth token is not set.')

        return False

    def _authenticate(self) -> None:
        logging.debug('About to authenticate with username {}.'.format(self._username))

        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        }

        credentials = {
            'username': self._username,
            'password': self._password,
        }

        request = requests.Request(method='POST', url=self.AUTH_URL, data=json.dumps(credentials), headers=headers)
        prepared_request = request.prepare()
        self._preprocess_request(prepared_request)

        response = requests.Session().send(prepared_request)
        self._preprocess_response(response)

        token = response.json()['accessToken']

        logging.debug('Successfully authenticated. Received token {}.'.format(token))

        token_padded = token + '=' * ((4 - len(token) % 4) % 4)
        self._token = token_padded



import logging
import re
import os
from typing import Match

import requests


class CZDSHelpers(object):

    def _is_email_address(self, address_string: str) -> Match:
        pat = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
        return re.match(pat, address_string)

    def _preprocess_request(self, request: requests.Request) -> None:
        logging.debug(
            'About to send {} request to URL {} with headers {} and body {}.'.format(
                request.method, request.url, request.headers, str(request.body)
            )
        )

    def _preprocess_response(self, response: requests.Response, stream: bool = False) -> None:
        logging.debug('Returned status code {}.'.format(response.status_code))
        logging.debug('Received headers {}.'.format(response.headers))

        if not stream:
            logging.debug('Received raw response {}.'.format(response.text))

        if response.status_code != 200:
            logging.error('Error during request to API. Returned status code {}.'.format(response.status_code))
            if response.status_code == 400:
                raise requests.HTTPError('Request is malformed.')
            if response.status_code == 401:
                raise requests.HTTPError('Invalid credentials provided, or no valid bearer token provided.')
            if response.status_code == 403:
                raise requests.HTTPError(
                    'Requested zone does not exist, or authenticated user does not have access to the zone.'
                )
            if response.status_code == 409:
                raise requests.HTTPError(
                    'The authenticated user has not accepted the new T&Cs. Please log in to the portal, and accept the T&Cs.'
                )
            if response.status_code == 415:
                raise requests.HTTPError('Unsupported content type header sent.')
            if response.status_code == 429:
                raise requests.HTTPError('Too many authentication attempts from the same IP address.')
            if response.status_code == 500:
                raise requests.HTTPError('Internal service error.')

    def _check_file_size(self, file_path: str, size: int) -> None:
        file_size = os.path.getsize(file_path)

        if file_size != size:
            raise Exception(
                'The size of the file ({:,} bytes) differs from the size announced in the header ({:,} bytes).'.format(
                    file_size, size
                )
            )



import unittest
import logging

from src.pyczds.client import CZDSClient


class TestPyCZDS(unittest.TestCase):
    def setUp(self):
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        # In order to run these tests, enter your valid username and password for the CZDS website here.
        # Please note that these tests do not change any data.

        # TODO PASTE USERNAME AND PASSWORD HERE
        username = ''
        password = ''
        # END TODO

        if not username or not password:
            raise Exception('Please set username and password in tests/test_pyczds.py before running it.')

        self.username = username
        self.password = password

        self.client = CZDSClient(username=username, password=password)
        self.client.USER_AGENT = 'pyCZDS / test'

        # the following is a token that is formally valid, but does not grant any access as it's expird.
        self.test_token = 'eyJraWQiOiJzZEF5bzA4TGpnZU54LWZzcGl3Q3VNTEE5RXRLRmlyX3g2OUFGeEQ5aUtZIiwiYWxnIjoiUlMyNTYifQ.eyJ2ZXIiOjEsImp0aSI6IkFULmRIWUw0X1JleTZ1QlE1azUwaTJOYU1Tbk81Tjlkc1pnUExRRU5UQTAya2siLCJpc3MiOiJodHRwczovL2ljYW5uLWFjY291bnQub2t0YS5jb20vb2F1dGgyL2F1czJwMDFjMnJvSkFlQ2dZMnA3IiwiYXVkIjoiaHR0cDovL2FwaV9hdXRoZW5yaXphdGlvbl9zZXJ2ZXIuaWNhbm4ub3JnIiwiaWF0IjoxNjcwNzkzNDc1LCJleHAiOjE2NzA4Nzk4NzUsImNpZCI6IjBvYTFyY2prcWtPbGlNUHVMMnA3IiwidWlkIjoiMDB1N2N5MXJqaktGbG01cW4ycDciLCJzY3AiOlsiaWNhbm4tY3VzdG9tIiwib3BlbmlkIl0sImF1dGhfdGltZSI6MTY3MDc5MzQ3NSwic3ViIjoidGVzdC1weWNkenNAZXhhbXBsZS5jb20iLCJnaXZlbl9uYW1lIjoiVGVzdCIsImZhbWlseV9uYW1lIjoiUHlDRFpTIiwiZW1haWwiOiJ0ZXN0LXB5Y2R6c0BleGFtcGxlLmNvbSJ9.A8_qsWDBy0vJGVv2a3Q64JTDYzxicdL0BdNsrl4dnQmYWhmwrJ7aAyNbmvWtBfd9_XhSswJUwZl9B61ANW-xpu5TSfUGF0R_2AWQ2ppzB2UUNUkCyetnfzd3LVwDXb6vqoEQLRAuVgAgxEOpwdCKlqNgsujWRg2e38w9qCtRAaiFTO8BvZ0eA_vo2zfWGnR6lo27J9ElWTUejZWBC1wzzwN60pZyjT3yJDfjBo4b8plXjKgtibyErWUdpOyCkhHbN6UFEyvfxyril6R8AZ8bWS5lzTmlvw_2yd9TG1xZA7hE9qbQJVK7ymslsOi7LFDSglcVa1vc6V0r1SCDH_k4Ig'



import random
import logging
import os
import tempfile
from email.message import EmailMessage
from datetime import datetime

from test_pyczds import TestPyCZDS


class TestClient(TestPyCZDS):

    # region online tests
    def test_get_zonefiles_list_online(self):
        zonefile_url_list = self.client.get_zonefiles_list()
        self.assertTrue(isinstance(zonefile_url_list, list))

    def test_head_zonefile_online(self):
        zonefile_url_list = self.client.get_zonefiles_list()
        url = random.choice(zonefile_url_list)

        head = self.client.head_zonefile(url)

        self.assertGreater(len(head), 0)
        self.assertIn('date', head)
        self.assertIn('server', head)
        self.assertIn('last-modified', head)
        self.assertGreater(int(head['content-length']), 0)
        self.assertEqual(head['content-type'], 'application/x-gzip')

    def test_parse_headers_online(self):
        zonefile_url_list = self.client.get_zonefiles_list()
        url = random.choice(zonefile_url_list)

        head = self.client.head_zonefile(url)

        self.assertIn('parsed', head)
        self.assertIsInstance(head['parsed']['last-modified'], datetime)
        self.assertIsInstance(head['parsed']['content-length'], int)
        self.assertIsInstance(head['parsed']['filename'], str)

        print(head)

    def download_smallest_zonefile(self, download_dir='', filename=''):
        zonefile_url_list = self.client.get_zonefiles_list()
        smallest_file_size = int()
        smallest_file_url = str()
        smallest_file_filename = str()

        for url in zonefile_url_list:
            head = self.client.head_zonefile(url)
            if smallest_file_size == 0 or int(head['Content-Length']) < smallest_file_size:
                smallest_file_size = int(head['Content-Length'])
                smallest_file_url = url

                msg = EmailMessage()
                msg['Content-Disposition'] = head['Content-Disposition']
                smallest_file_filename = msg.get_filename()

        logging.debug(
            'Identified {} as smallest file with {} bytes to download.'.format(smallest_file_url, smallest_file_size)
        )

        if len(filename) > 0:
            local_filename = filename
        else:
            local_filename = smallest_file_filename

        self.client.get_zonefile(smallest_file_url, download_dir, local_filename)

        if len(download_dir) > 0:
            filepath = os.path.join(download_dir, local_filename)
        else:
            filepath = os.path.join(os.getcwd(), local_filename)

        filesize = os.stat(filepath)

        self.assertEqual(smallest_file_size, filesize.st_size)

        logging.debug('Removing downloaded file {} from {}.'.format(local_filename, filepath))

        os.remove(filepath)

    def test_download_smallest_zonefile_default_directory_online(self):
        self.download_smallest_zonefile()

    def test_download_smallest_zonefile_other_filename_online(self):
        with tempfile.TemporaryDirectory() as d:
            self.download_smallest_zonefile(filename='test_zonefile.tmp')

    def test_download_smallest_zonefile_other_directory_online(self):
        with tempfile.TemporaryDirectory() as d:
            self.download_smallest_zonefile(d)

    def test_download_smallest_zonefile_other_directory_other_filename_online(self):
        with tempfile.TemporaryDirectory() as d:
            self.download_smallest_zonefile(d, 'test_zonefile.tmp')

    # endregion



import base64
import json

from requests import HTTPError

from src.pyczds.client import CZDSClient
from test_pyczds import TestPyCZDS


class TestAuthentication(TestPyCZDS):
    # region offline tests
    def test_instantiate_client_no_username_offline(self):
        # Client should raise ValueError if username is empty.
        with self.assertRaises(ValueError):
            CZDSClient(username='', password='password')

    def test_instantiate_client_invalid_username_offline(self):
        # Client should raise ValueError if username is not an email address.
        with self.assertRaises(ValueError):
            CZDSClient(username='user', password='password')

    def test_instantiate_client_no_password_offline(self):
        # Client should raise ValueError if password is empty.
        with self.assertRaises(ValueError):
            client = CZDSClient(username='test@test.com', password='')

    def test_jwt_payload_returned_offline(self):
        # Client should return the payload in the JWT portion of the token.
        self.client._token = self.test_token
        j = self.client._get_token_jwt_payload()

        payload_b64 = self.test_token.split('.')[1]
        s = base64.b64decode(payload_b64)
        j_reference = json.loads(s)

        self.assertEqual(j, j_reference)

    def test_jwt_payload_raises_empty_token_offline(self):
        # If token is empty, method should raise a ValueError.
        self.client._token = str()
        with self.assertRaises(ValueError):
            j = self.client._get_token_jwt_payload()

    def test_not_authenticated_token_empty_offline(self):
        # Client should not be authenticated if there's no token saved.
        self.client._token = ''
        self.assertFalse(self.client._is_authenticated())

    def test_not_authenticated_token_expired_offline(self):
        # Client should not be authenticated if token has expired.
        self.client._token = self.test_token
        self.assertFalse(self.client._is_authenticated())
    # endregion

    # region online tests
    def test_authenticate_invalid_username_online(self):
        # Client should raise an HTTPError if authentication is unsuccessful.
        self.client._username = 'invalid_user_pycdzs'
        self.client._password = 'invalid_password_pycdzs'

        with self.assertRaisesRegex(HTTPError, '.*nvalid credentials provided.*'):
            self.client._authenticate()

    def test_authenticate_invalid_password_online(self):
        # Client should raise an HTTPError if authentication is unsuccessful.
        self.client._username = self.username
        self.client._password = 'invalid_password_pycdzs'

        # There seems to be a bug in the API. If the username is correct, but the password is wrong,
        # instead of status code 401 (Invalid credentials), status code 400 ("Request is malformed") is returns.
        with self.assertRaisesRegex(HTTPError, '.*equest.*malformed.*'):
            self.client._authenticate()

    def test_authenticate_valid_username_password_online(self):
        # Client should be authenticated if username and password are correct.
        self.client._username = self.username
        self.client._password = self.password

        self.client._authenticate()
        self.assertTrue(self.client._is_authenticated())

    def test_jwt_payload_returned_online(self):
        # Client should return the payload in the JWT portion of the token.
        self.client._authenticate()
        j = self.client._get_token_jwt_payload()
        self.assertEqual(j['sub'], self.client._username)

    # endregion
